use clap::Parser;
use readmap::types::{Content::*, Section};
use tool_helper::{Error, exit_with_error};
use std::{fs::File, io::{BufRead, BufReader, BufWriter, Write}, path::PathBuf, process::{Child, Command, Stdio}};

#[derive(Parser)]
#[clap(about = "Opens and scans a MAP file to print extended information", long_about = None)]
struct CommandLine {
    #[clap(value_parser, help="Input MAP file for scannning")]
    input:          PathBuf,
    #[clap(long="wsl", default_value_t=false)]
    use_wsl:        bool,
}

fn run_objdump(use_wsl: bool, input: PathBuf) -> Result<Child, Error> {
    let command_list = ["wsl", "objdump", "-x"];
    let start_idx    = {
        if use_wsl {
            0
        }

        else {
            1
        }
    };
    let mut process = Command::new(command_list[start_idx]);

    for arg in command_list.iter().skip(start_idx + 1) {
        process.arg(arg);
    }
    process.arg(input.to_str().ok_or_else(||{Error::from_str("Failed converting input string")})?);

    println!("Miau: {:?}", process);
    Ok(process.stdout(Stdio::piped()).spawn()?)
}

fn run_main(cmd: CommandLine) -> Result<(), Error> {
    let mut child = run_objdump(cmd.use_wsl, cmd.input)?;
    if let Some(stdout) = &mut child.stdout {
        for line in BufReader::new(stdout).lines() {
            if let Ok(line) = line {
                println!("{}", line);
            }
        }
    }

    Ok(())
}

fn _run_main(cmd: CommandLine) -> Result<(), Error> {
    fn value_to_hex<T: std::fmt::UpperHex>(value: T) -> String {
        return format!("0x{:X}", value);
    }

    fn option_to_hex<T: std::fmt::UpperHex>(value: Option<T>) -> String {
        if let Some(value) = value {
            return value_to_hex(value);
        }

        else {
            return String::from("<None>");
        }
    }

    let     sections = readmap::scan(tool_helper::open_input(Some(cmd.input))?)?;
    let mut file     = tool_helper::open_output_file(&PathBuf::from("./planschi.d"))?;

    for section in sections {
        fn print_content(tab_level: usize, file: &mut BufWriter<File>, section: Section) -> Result<(), Error> {
            for content in section.content {
                match content {
                    Fill(fill)       => {
                        writeln!(file, "{:>tab_level$}*fill* @{}, {}", ' ', value_to_hex(fill.adr), value_to_hex(fill.size), tab_level=tab_level*4)?;
                    },
                    Section(section) => {
                        writeln!(file, "{:>tab_level$} {} @{}, {}", ' ', section.name, option_to_hex(section.adr), option_to_hex(section.size), tab_level=tab_level*4)?;
                        print_content(tab_level + 1, file, section)?;
                    },
                    Symbol(symbol)   => {
                        writeln!(file, "{:>tab_level$}{} @{}", ' ', symbol.name, value_to_hex(symbol.adr), tab_level=tab_level*4)?;
                    }
                }
            }

            Ok(())
        }

        writeln!(file, "{}: @{}, {}", section.name, option_to_hex(section.adr), option_to_hex(section.size))?;
        print_content(1, &mut file, section)?;
    }
    Ok(())
}

pub fn main() {
    match CommandLine::try_parse() {
        Ok(cmd_line) => {
            match run_main(cmd_line) {
                Ok(_)      => (),
                Err(error) => exit_with_error(error)
            }
        },
        Err(error) => {
            println!("{}", error)
        }
    }
}